#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import asyncio
import re


__all__ = ["DepthGauges"]


class DepthGauges:
    """Reads the value of Heidenhain depth gauges."""

    def __init__(self, host: str, port: int, camera: str | None = None):

        self.host = host
        self.port = port
        self.camera = camera

    async def read(self):
        """Returns the measured values from the depth probes."""

        depth = {"A": -999.0, "B": -999.0, "C": -999.0}

        for channel in depth:
            w = None
            try:
                conn = asyncio.open_connection(self.host, self.port)
                r, w = await asyncio.wait_for(conn, 1)
                w.write(("SEND " + channel + "\n").encode())
                await w.drain()
                reply = await asyncio.wait_for(r.readline(), 1)
            except Exception:
                raise ValueError("Failed retrieving data from depth probes.")
            finally:
                if w is not None:
                    w.close()
                    await w.wait_closed()

            match = re.match(f"\r?{channel} ([+\\-0-9\\.]+) mm".encode(), reply)
            if match:
                depth[channel] = float(match.group(1).decode())
            else:
                raise ValueError(f"Failed parsing depth probe for channel {channel}")

        return depth



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import asyncio
import re
import warnings
from dataclasses import dataclass

from lvmieb.exceptions import LvmIebUserWarning


__all__ = ["PressureTransducer"]


@dataclass
class PressureTransducer:
    """Communicates with a SENS4 pressure transducer.

    Parameters
    ----------
    spec
        The spectrograph associated with this device.
    camera
        The camera (red, blue, NIR) that this device monitors.
    host
        The host that provides the TCP server.
    port
        The port on which the TCP server is running.
    device_id
        The ID of the device.

    """

    spec: str
    camera: str
    host: str
    port: int
    device_id: int = 254

    async def _read(self, query_string: str = "P"):
        """Queries the transducer."""

        try:
            r, w = await asyncio.wait_for(
                asyncio.open_connection(self.host, self.port), 1
            )
        except Exception as err:
            warnings.warn(
                f"Transducer {self.camera}: failed connecting to host: {err}",
                LvmIebUserWarning,
            )
            return False

        command = "@" + str(self.device_id) + query_string + "?\\"
        w.write(command.encode())
        await w.drain()

        try:
            reply = await asyncio.wait_for(r.readuntil(b"\\"), 1)
            match = re.search(r"@[0-9]{1,3}ACK([0-9.E+-]+)\\$".encode(), reply)
            if not match:
                raise ValueError("Cannot parse reply.")
            return float(match.groups()[0])
        except Exception as err:
            warnings.warn(
                f"Transducer {self.camera}: query {command!r} failed: {err}",
                LvmIebUserWarning,
            )
            return False
        finally:
            w.close()
            await w.wait_closed()

    async def read_pressure(self):
        """Reads the pressure from the transducer."""

        return await self._read("P")

    async def read_temperature(self):
        """Reads the temperature from the transducer."""

        return await self._read("T")



# -*- coding: utf-8 -*-
#
# @Author: Changgon Kim, Mingyeong YANG, Taeeun Kim
# @Date: 2021-03-22
# @Filename: telemetry.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import click

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor.actor import ControllersType, IEBCommand


@parser.group()
def wago(*args):
    """Controls the WAGO IOModule."""
    pass


@wago.command()
@click.argument("SPECTRO", type=click.Choice(["sp1", "sp2", "sp3"]), required=False)
async def status(
    command: IEBCommand,
    controllers: ControllersType,
    spectro: str | None = None,
):
    """Returns the status of WAGO sensors."""

    spectro_list: list[str] = []
    if spectro is None:
        spectro_list = [controller_name for controller_name in controllers]
    else:
        spectro_list = [spectro]

    tasks = []
    for spectro_name in spectro_list:
        if spectro_name not in controllers:
            command.warning(error=f"Spectrograph {spectro!r} is not available.")
            continue

        controller = controllers[spectro_name]
        tasks.append(controller.wago.read_sensors())

    results = await asyncio.gather(*tasks, return_exceptions=True)
    if len(results) == 0:
        return command.fail("No data received.")

    sensors = {}
    for ii, spectro_name in enumerate(spectro_list):
        result = results[ii]
        if isinstance(result, Exception):
            command.warning(f"Failed to read {spectro_name} sensors.")
            continue
        sensors[f"{spectro_name}_sensors"] = result

    return command.finish(**sensors)


@wago.command()
@click.argument("SPECTRO", type=click.Choice(["sp1", "sp2", "sp3"]), required=False)
async def getpower(
    command: IEBCommand,
    controllers: ControllersType,
    spectro: str | None = None,
):
    """Returns the status of the power relays."""

    spectro_list: list[str] = []
    if spectro is None:
        spectro_list = [controller_name for controller_name in controllers]
    else:
        spectro_list = [spectro]

    tasks = []
    for spectro_name in spectro_list:
        if spectro_name not in controllers:
            command.warning(error=f"Spectrograph {spectro!r} is not available.")
            continue

        controller = controllers[spectro_name]
        tasks.append(controller.wago.read_relays())

    results = await asyncio.gather(*tasks, return_exceptions=True)
    if len(results) == 0:
        return command.fail("No data received.")

    relays = {}
    for ii, spectro_name in enumerate(spectro_list):
        result = results[ii]
        if isinstance(result, Exception):
            command.warning(f"Failed to read {spectro_name} relays.")
            continue
        relays[f"{spectro_name}_relays"] = result

    return command.finish(**relays)


@wago.command()
@click.argument(
    "DEVICE",
    type=click.Choice(["shutter", "hartmann_left", "hartmann_right"]),
)
@click.argument("SPECTRO", type=click.Choice(["sp1", "sp2", "sp3"]))
@click.option(
    "--on",
    "action",
    flag_value="ON",
    required=True,
    help="Turn on device",
)
@click.option(
    "--off",
    "action",
    flag_value="OFF",
    required=True,
    help="Turn off device",
)
async def setpower(
    command: IEBCommand,
    controllers: ControllersType,
    device: str,
    action: str,
    spectro: str,
):
    """Switches power on/off to a device."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    try:
        closed = True if action == "ON" else False
        await controller.wago.set_relay(device, closed=closed)
    except NameError:
        return command.fail(error=f"Cannot find device {device!r}.")

    await (await command.child_command(f"wago getpower {spectro}"))

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import asyncio
import re
import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from lvmieb.controller.maskbits import MotorStatus
from lvmieb.exceptions import LvmIebUserWarning, MotorControllerError


if TYPE_CHECKING:
    from lvmieb.controller.wago import IEBWAGO


__all__ = ["DEVLIST", "MotorController", "parse_IS"]

# Device list
DEVLIST = ["shutter", "hartmann_left", "hartmann_right"]


# Shutter/HD commands
COMMANDS = {"init": "QX1", "home": "QX2", "open": "QX3", "close": "QX4", "status": "IS"}


@dataclass
class MotorController:
    """Handles a motor controller for the shutter or Hartmann doors.

    Parameters
    ----------
    spec
        The spectrograph with which this motor controller is associated.
    type
        The type of motor controller, one of `.DEVLIST`. Although all motor
        controllers have the same command set, the proximity sensor status
        depends on the type of device.
    host
        The host providing the TCP server.
    port
        The port to the TCP server.
    wago
        Optionally, the `.IEBWAGO` instance associated with this device, used
        to check the power status of the motor controller.

    """

    spec: str
    type: str
    host: str
    port: int
    wago: Optional[IEBWAGO] = None

    TIMEOUT: float = 5

    def __post_init__(self):

        if self.type not in DEVLIST:
            raise ValueError(f"Device type {self.type} is not valid.")

    async def get_power_status(self):
        """Returns the power status of a motor controller."""

        if self.wago is None:
            raise RuntimeError("WAGO module not specified.")

        device = self.wago.get_device(self.type)
        return (await device.read())[0] == "closed"

    async def send_command(self, command: str, timeout: float = 1) -> bytes:
        """Sends a command to the device."""

        if command in COMMANDS:
            command = COMMANDS[command]

        try:
            conn = asyncio.open_connection(self.host, self.port)
            r, w = await asyncio.wait_for(conn, self.TIMEOUT)
        except OSError as err:
            raise MotorControllerError(
                f"{self.type} ({self.spec}): failed connecting to device: {err}"
            )
        except asyncio.TimeoutError:
            raise MotorControllerError(
                f"{self.type} ({self.spec}): timed out connecting to device."
            )

        w.write((f"\00\07{command}\r").encode())
        await w.drain()

        reply = b""
        try:
            while True:
                reply += await asyncio.wait_for(r.readuntil(b"\r"), timeout)
                if command == "IS":
                    return reply
                if b"ERR" in reply or b"DONE" in reply:
                    return reply
        except asyncio.TimeoutError:
            raise MotorControllerError(
                f"{self.type} ({self.spec}): timed out waiting for "
                f"reply to {command!r}.",
            )
        finally:
            w.close()
            await w.wait_closed()

    async def get_status(self) -> tuple[MotorStatus, str | None]:
        """Returns the status and position of the motor.

        This method never raises an error; connection issues or invalid
        states are encoded as `.MotorStatus` bits.

        Returns
        -------
        status
            A tuple in which the first element is a `.MotorController` flag with
            the full status of the motor, and the second is the bytestring returned
            by the ``IS`` command. The second element is null if the command fails.

        """

        motor_status = MotorStatus.POWER_UNKNOWN

        if self.wago is not None:
            power = await self.get_power_status()
            if power is False:
                motor_status = MotorStatus.POWER_OFF | MotorStatus.POSITION_UNKNOWN
                return (motor_status, None)
            else:
                motor_status = MotorStatus.POWER_ON

        try:
            reply = await self.send_command("status")
        except MotorControllerError as err:
            warnings.warn(str(err), LvmIebUserWarning)
            motor_status |= MotorStatus.POSITION_UNKNOWN
            return (motor_status, None)

        match = re.search(b"\x00\x07IS=([0-1]{8})\r$", reply)
        if match is None:
            warnings.warn(
                f"Cannot match reply {reply} for {self.type} in {self.spec}",
                LvmIebUserWarning,
            )
            motor_status |= MotorStatus.POSITION_INVALID
            return (motor_status, None)

        bits = match.group(1).decode()
        status = parse_IS(reply, self.type)

        if status == "open":
            motor_status |= MotorStatus.OPEN
        elif status == "closed":
            motor_status |= MotorStatus.CLOSED
        else:
            motor_status |= MotorStatus.POSITION_INVALID

        return (motor_status, bits)

    async def move(self, open: bool | None = None, force: bool = False) -> bool:
        """Moves the device.

        Parameters
        ----------
        open
            If `True`, opens the device, if `False`, closes it. With `None`,
            switches the position of the device.
        force
            Send the command even if the device is already at the position.

        Returns
        -------
        result
            `True` if the command succeeded or the device is already at the
            position, `False` if it failed.

        """

        status = (await self.get_status())[0]

        if status & (MotorStatus.POSITION_UNKNOWN | MotorStatus.POSITION_INVALID):
            raise MotorControllerError("Motor position is unknown or invalid.")

        if open is None:
            if status & MotorStatus.OPEN:
                command = "close"
            elif status & MotorStatus.CLOSED:
                command = "open"
            else:
                raise ValueError("Invalid motor status.")
        elif open is True:
            command = "open" if (status & MotorStatus.CLOSED or force) else None
        elif open is False:
            command = "close" if (status & MotorStatus.OPEN or force) else None
        else:
            raise ValueError(f"Invalid motor status {open!r}.")

        if command is None:
            return True

        reply = await self.send_command(command, timeout=3.0)

        if b"DONE" in reply:
            return True
        elif b"ERR" in reply:
            return False
        else:
            raise MotorControllerError(
                f"{self.type} ({self.spec}): invalid reply to command {command!r}."
            )


def parse_IS(reply: bytes, device: str):
    """Parses the reply to the shutter IS command."""

    match = re.search(b"\x00\x07IS=([0-1])([0-1])[0-1]{6}\r$", reply)
    if match is None:
        return None

    if match.groups() == (b"1", b"0"):
        if device in ["shutter", "hartmann_right"]:
            return "open"
        else:
            return "closed"
    elif match.groups() == (b"0", b"1"):
        if device in ["shutter", "hartmann_right"]:
            return "closed"
        else:
            return "open"
    else:
        return None



# -*- coding: utf-8 -*-
#
# @Author: Changging Kim, Mingyeong Yang, Taeeun Kim
# @Date: 2020-10-26
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os

import click
from click_default_group import DefaultGroup

from clu.tools import cli_coro as cli_coro_lvm
from sdsstools.daemonizer import DaemonGroup

from lvmieb.actor.actor import IEBActor


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode. Use additional v for more details.",
)
@click.pass_context
def lvmieb(ctx, config_file, verbose):
    """LVM Electronics Box Controller."""

    ctx.obj = {"verbose": verbose, "config_file": config_file}


@lvmieb.group(cls=DaemonGroup, prog="lvmieb_actor", workdir=os.getcwd())
@click.pass_context
@cli_coro_lvm
async def actor(ctx):
    """Runs the actor."""

    default_config_file = os.path.join(os.path.dirname(__file__), "etc/lvmieb.yml")
    config_file = ctx.obj["config_file"] or default_config_file

    lvmieb_obj = IEBActor.from_config(config_file)

    if ctx.obj["verbose"]:
        lvmieb_obj.log.fh.setLevel(0)
        lvmieb_obj.log.sh.setLevel(0)

    await lvmieb_obj.start()
    await lvmieb_obj.run_forever()


if __name__ == "__main__":
    lvmieb()



# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Changgon Kim, Mingeyong Yang, Taeeun Kim
# @Date:    2021-04-26 17:14
# @Last modified by:   Changgon Kim

from __future__ import annotations


class LvmIebError(Exception):
    """A custom core LvmIeb exception"""

    def __init__(self, message=None):

        message = "There has been an error" if not message else message

        super(LvmIebError, self).__init__(message)


class MotorControllerError(LvmIebError):
    """A motor controller error."""


class LvmIebWarning(Warning):
    """Base warning for LvmIeb."""


class LvmIebUserWarning(UserWarning, LvmIebWarning):
    """The primary warning class."""

    pass


class LvmIebSkippedTestWarning(LvmIebUserWarning):
    """A warning for when a test is skipped."""

    pass


class LvmIebDeprecationWarning(LvmIebUserWarning):
    """A warning for deprecated features."""

    pass



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import click

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor.actor import ControllersType, IEBCommand


@parser.group()
def transducer(*args):
    """Reports pressure transducer values.."""
    pass


@transducer.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]), required=False)
async def status(
    command: IEBCommand,
    controllers: ControllersType,
    spectro: str | None = None,
):
    """Returns the status of transducer."""

    tasks = []

    pres_result = {}
    camera_order = []

    for controller_name in controllers:
        if spectro is not None and controller_name != spectro:
            continue
        controller = controllers[controller_name]
        for camera, pressure_transducer in controller.pressure.items():
            tasks.append(pressure_transducer.read_pressure())
            tasks.append(pressure_transducer.read_temperature())
            camera_order.append(camera)

    pres_list = await asyncio.gather(*tasks, return_exceptions=True)

    for ii, camera in enumerate(camera_order):
        camera_results = pres_list[2 * ii : 2 * ii + 2]
        for jj, measurement in enumerate(["pressure", "temperature"]):
            camera_result = camera_results[jj]
            if isinstance(camera_result, Exception):
                command.warning(f"Failed getting {camera} {measurement}.")
                camera_result = -999.0
            pres_result[f"{camera}_{measurement}"] = camera_result

    command.finish(transducer=pres_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import enum


__all__ = ["MotorStatus"]


class MotorStatus(enum.Flag):
    """Status bits for motor status."""

    POWER_ON = 0x1
    POWER_OFF = 0x2
    POWER_UNKNOWN = 0x4
    OPEN = 0x10
    CLOSED = 0x20
    POSITION_INVALID = 0x40
    POSITION_UNKNOWN = 0x80



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Changgon Kim, Taeeun Kim, Mingyeong YANG (mingyeong@khu.ac.kr)
# @Date: 2021-03-22
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import pathlib
from copy import deepcopy
from typing import ClassVar

from yaml import warnings

from clu import Command
from clu.actor import AMQPActor
from sdsstools.configuration import read_yaml_file

from lvmieb import __version__, config
from lvmieb.controller.controller import IEBController
from lvmieb.controller.depth import DepthGauges
from lvmieb.exceptions import LvmIebUserWarning

from .commands import parser as lvm_command_parser


__all__ = ["IEBActor", "IEBCommand", "ControllersType"]


class IEBActor(AMQPActor):
    """LVM Spectrograph controller actor.

    In addition to the normal arguments and keyword parameters for
    `~clu.actor.AMQPActor`, the class accepts the following parameters.

    Parameters
    ----------
    controllers
        The list of '.IEBController' instances to manage.

    """

    parser = lvm_command_parser
    BASE_CONFIG: ClassVar[str | dict | None] = config

    def __init__(
        self,
        *args,
        controllers: tuple[IEBController, ...] = (),
        depth_gauges: DepthGauges | None = None,
        **kwargs,
    ):

        self.controllers = {c.spec: c for c in controllers}
        self.depth_gauges = depth_gauges

        self.version = __version__

        super().__init__(*args, **kwargs)

    async def start(self, **kwargs):  # pragma: no cover
        """Starts the actor connection to RabbitMQ."""

        # We wait to set this until here to make sure all the controllers
        # have been created.
        self.parser_args = [self.controllers]

        return await super().start(**kwargs)

    @classmethod
    def from_config(cls, config: dict | str | None, *args, **kwargs):
        """Creates an actor from a configuration file."""

        if config is None:
            if cls.BASE_CONFIG is None:
                raise RuntimeError("The class does not have a base configuration.")
            config = cls.BASE_CONFIG

        if isinstance(config, (str, pathlib.Path)):
            config = dict(read_yaml_file(str(config)))

        if "schema" in config["actor"]:
            schema = config["actor"]["schema"]
            if not os.path.isabs(schema):
                schema = os.path.join(
                    os.path.abspath(os.path.join(os.path.dirname(__file__), "../")),
                    schema,
                )
            config["actor"]["schema"] = schema

        instance = super().from_config(config["actor"])

        controllers: list[IEBController] = []

        for spec in config.get("enabled_specs", []):
            if "specs" not in config or spec not in config["specs"]:
                warnings.warn(
                    f"Cannot find configuration for {spec!r}.",
                    LvmIebUserWarning,
                )
                continue

            spec_config = config["specs"][spec].copy()
            wago_modules = deepcopy(config.get("wago_modules", {}))

            controller = IEBController.from_config(
                spec,
                spec_config,
                wago_modules=wago_modules,
            )
            controllers.append(controller)

        instance.controllers = {contr.spec: contr for contr in controllers}

        if (depth_gauges := config.get("depth_gauges", None)) is not None:
            instance.depth_gauges = DepthGauges(**depth_gauges.copy())

        return instance


IEBCommand = Command[IEBActor]
ControllersType = dict[str, IEBController]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from lvmieb.controller.motor import MotorController
from lvmieb.controller.pressure import PressureTransducer
from lvmieb.controller.wago import IEBWAGO


class IEBController:
    """Talks to an Ieb controller over TCP/IP

    This class mainly combines the different subsystem controllers (WAGO,
    motors, transducers) under a common umbrella.

    """

    def __init__(
        self,
        spec: str,
        wago: IEBWAGO,
        pressure: list[PressureTransducer] = [],
        motors: list[MotorController] = [],
    ):

        self.spec = spec

        self.wago = wago
        self.pressure = {p.camera: p for p in pressure}
        self.motors = {m.type: m for m in motors}

    @classmethod
    def from_config(cls, spec: str, config: dict, wago_modules: dict = {}):
        """Creates an instance of `.IEBController` from a configuration file."""

        wago_config = config["wago"].copy()
        wago_config["modules"] = wago_modules.copy()

        wago = IEBWAGO.from_config(wago_config, name=spec)

        motors = []
        for motor, motor_config in config.get("motor_controllers", {}).copy().items():
            motors.append(MotorController(spec, motor, **motor_config, wago=wago))

        pressure = []
        for camera, pressure_config in config.get("pressure", {}).copy().items():
            pressure.append(PressureTransducer(spec, camera, **pressure_config))

        return cls(spec, wago, pressure=pressure, motors=motors)



# -*- coding: utf-8 -*-
# @Author: Changgon Kim, Mingyeong Yang, Taeeun Kim
# @Date: 2021-05-12
# @Filename: hartmann.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# added by CK 2021/03/30

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import click

from lvmieb.controller.maskbits import MotorStatus
from lvmieb.exceptions import MotorControllerError

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor import ControllersType, IEBCommand


__all__ = ["hartmann"]


@parser.group()
def hartmann(*args):
    """Control the hartmann doors."""

    pass


@hartmann.command()
@click.option(
    "-s",
    "--side",
    type=click.Choice(["all", "right", "left"]),
    default="all",
    help="all, right, or left",
)
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def open(
    command: IEBCommand,
    controllers: ControllersType,
    side: str,
    spectro: str,
):
    """Open the Hartmann doors."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    if side == "all" or side == "left":
        tasks.append(controller.motors["hartmann_left"].move(open=True))
    if side == "all" or side == "right":
        tasks.append(controller.motors["hartmann_right"].move(open=True))

    command.info(text=f"Opening {side} hartmanns")
    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    await (await command.child_command(f"hartmann status {spectro}"))

    return command.finish()


@hartmann.command()
@click.option(
    "-s",
    "--side",
    type=click.Choice(["all", "right", "left"]),
    default="all",
    help="all, right, or left",
)
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def close(
    command: IEBCommand,
    controllers: ControllersType,
    side: str,
    spectro: str,
):
    """Close the Hartmann doors."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    if side == "all" or side == "left":
        tasks.append(controller.motors["hartmann_left"].move(open=False))
    if side == "all" or side == "right":
        tasks.append(controller.motors["hartmann_right"].move(open=False))

    command.info(text=f"Closing {side} hartmanns")
    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    await (await command.child_command(f"hartmann status {spectro}"))

    return command.finish()


@hartmann.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def status(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Reports the position of the Hartmann doors."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    tasks.append(controller.motors["hartmann_left"].get_status())
    tasks.append(controller.motors["hartmann_right"].get_status())

    result_hartmann = await asyncio.gather(*tasks)

    hd_status = {}
    for i, name in enumerate(["left", "right"]):
        motor_status, bits = result_hartmann[i]

        power = motor_status & MotorStatus.POWER_ON
        open = motor_status & MotorStatus.OPEN
        invalid = motor_status & (
            MotorStatus.POSITION_INVALID
            | MotorStatus.POSITION_UNKNOWN
            | MotorStatus.POWER_UNKNOWN
        )

        hd_status[name] = {
            "power": power.value > 0,
            "open": open.value > 0,
            "invalid": invalid.value > 0,
            "bits": bits or "?",
        }

    command.info({f"{spectro}_hartmann_left": hd_status["left"]})
    command.info({f"{spectro}_hartmann_right": hd_status["right"]})

    return command.finish()


@hartmann.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def init(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Initialise the hartmanns."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    command.info(text="Initializing all hartmanns")

    tasks = []
    for hd in ["hartmann_left", "hartmann_right"]:
        tasks.append(controller.motors[hd].send_command("init"))

    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    return command.finish()


@hartmann.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def home(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Home the hartmann doors."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    command.info(text="Home all hartmanns")

    tasks = []
    for hd in ["hartmann_left", "hartmann_right"]:
        tasks.append(controller.motors[hd].send_command("home"))

    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    return command.finish()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python'}

packages = \
['lvmieb', 'lvmieb.actor', 'lvmieb.actor.commands', 'lvmieb.controller']

package_data = \
{'': ['*'], 'lvmieb': ['etc/*']}

install_requires = \
['click-default-group>=1.2.2,<2.0.0',
 'numpy>=1.20.3,<2.0.0',
 'sdss-clu>=1.6.1,<2.0.0',
 'sdss-drift>=0.4.2,<0.5.0',
 'sdsstools>=0.4.0']

entry_points = \
{'console_scripts': ['lvmieb = lvmieb.__main__:lvmieb']}

setup_kwargs = {
    'name': 'sdss-lvmieb',
    'version': '0.4.0',
    'description': 'Control software for the Local Volume Mapper Instrument Electronics Box',
    'long_description': '# lvmieb\n\n![Versions](https://img.shields.io/badge/python->3.8-blue)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/lvmieb/badge/?version=latest)](https://lvmieb.readthedocs.io/en/latest/?badge=latest)\n[![Test](https://github.com/sdss/lvmieb/actions/workflows/test.yml/badge.svg)](https://github.com/sdss/lvmieb/actions/workflows/test.yml)\n[![Docker](https://github.com/sdss/lvmieb/actions/workflows/docker.yml/badge.svg)](https://github.com/sdss/lvmieb/actions/workflows/docker.yml)\n[![codecov](https://codecov.io/gh/sdss/lvmieb/branch/main/graph/badge.svg?token=IyQglaQYSF)](https://codecov.io/gh/sdss/lvmieb)\n\nControl software for the SDSS-V LVM (Local Volume Mapper) spectrograph Instrument Electronics Box (IEB).\n\n## Quick Start\n\n### Installation\n\n`lvmieb` uses the [CLU](https://clu.readthedocs.io/en/latest/) framework and requires a RabbitMQ instance running in the background.\n\n`lvmieb` can be installed using `pip`\n\n```console\npip install sdss-lvmieb\n```\n\nor by cloning this repository\n\n```console\ngit clone https://github.com/sdss/lvmieb\n```\n\nThe preferred installation for development is using [poetry](https://python-poetry.org/)\n\n```console\ncd lvmieb\npoetry install\n```\n\n\n### Basic ping-pong test\n\nStart the `lvmieb` actor.\n\n```console\nlvmieb start\n```\n\nIn another terminal, type `clu` and `lvmieb ping` for test.\n\n```console\n$ clu\nlvmieb ping\n07:41:22.636 lvmieb >\n07:41:22.645 lvmieb : {\n    "text": "Pong."\n}\n```\n\nStop `lvmieb` actor.\n\n```console\nlvmieb stop\n```\n',
    'author': 'Changgon Kim',
    'author_email': 'changgonkim@khu.ac.kr',
    'maintainer': 'José Sánchez-Gallego',
    'maintainer_email': 'gallegoj@uw.edu',
    'url': 'https://github.com/sdss/lvmieb',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor import ControllersType, IEBCommand


__all__ = ["depth"]


@parser.group()
def depth(*args):
    """Controls the linear gauge depth."""
    pass


@depth.command()
@click.option(
    "--camera",
    type=str,
    help="Temporarily sets the camera to which the probes are connected.",
)
async def status(
    command: IEBCommand,
    controllers: ControllersType,
    camera: str | None = None,
):
    """Returns the measurements from the depth probes."""

    depth_gauges = command.actor.depth_gauges
    if depth_gauges is None:
        return command.fail(error="Depth gauge configuration not defined.")

    try:
        depth = await depth_gauges.read()
    except Exception as err:
        return command.fail(error=err)

    camera = camera or depth_gauges.camera or "?"
    return command.finish({"depth": {"camera": camera, **depth}})


@depth.command(name="set-camera")
@click.argument("CAMERA", type=str)
async def set_camera(command: IEBCommand, controllers: ControllersType, camera: str):
    """Sets the camera to which the probes are connected."""

    depth_gauges = command.actor.depth_gauges
    if depth_gauges is None:
        return command.fail(error="Depth gauge configuration not defined.")

    depth_gauges.camera = camera

    return command.finish()



# -*- coding: utf-8 -*-
#
# @Author: Changgon Kim, Mingyeong Yang, Taeeun Kim
# @Date: 2021-05-12
# @Filename: shutter.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
# added by CK 2021/03/30

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import click

from lvmieb.controller.maskbits import MotorStatus
from lvmieb.exceptions import MotorControllerError

from . import parser


if TYPE_CHECKING:
    from lvmieb.actor import ControllersType, IEBCommand


__all__ = ["shutter"]


@parser.group()
def shutter(*args):
    """Control the shutter."""

    pass


@shutter.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def open(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Open the shutter."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    tasks.append(controller.motors["shutter"].move(open=True))

    command.info(text="Opening shutter")
    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    await (await command.child_command(f"shutter status {spectro}"))

    return command.finish()


@shutter.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def close(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Close the shutter."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    tasks.append(controller.motors["shutter"].move(open=False))

    command.info(text="Closing shutter")
    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    await (await command.child_command(f"shutter status {spectro}"))

    return command.finish()


@shutter.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def status(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Reports the position of the shutter."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    tasks = []
    tasks.append(controller.motors["shutter"].get_status())

    result_shutter = await asyncio.gather(*tasks)

    shutter_status = {}
    motor_status, bits = result_shutter[0]

    power = motor_status & MotorStatus.POWER_ON
    open = motor_status & MotorStatus.OPEN
    invalid = motor_status & (
        MotorStatus.POSITION_INVALID
        | MotorStatus.POSITION_UNKNOWN
        | MotorStatus.POWER_UNKNOWN
    )

    shutter_status = {
        "power": power.value > 0,
        "open": open.value > 0,
        "invalid": invalid.value > 0,
        "bits": bits or "?",
    }

    return command.finish({f"{spectro}_shutter": shutter_status})


@shutter.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def init(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Initialise the shutter."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    command.info(text="Initializing shutter")

    tasks = []
    tasks.append(controller.motors["shutter"].send_command("init"))

    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    return command.finish()


@shutter.command()
@click.argument("spectro", type=click.Choice(["sp1", "sp2", "sp3"]))
async def home(command: IEBCommand, controllers: ControllersType, spectro: str):
    """Home the shutter."""

    if spectro not in controllers:
        return command.fail(error=f"Spectrograph {spectro!r} is not available.")

    controller = controllers[spectro]

    command.info(text="Homing shutter")

    tasks = []
    tasks.append(controller.motors["shutter"].send_command("home"))

    try:
        await asyncio.gather(*tasks)
    except MotorControllerError as err:
        return command.fail(error=err)

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from drift import Drift, Relay


__all__ = ["IEBWAGO"]


class IEBWAGO(Drift):
    """Class controlling the WAGO PLC in each electronics box.

    Parameters
    ----------
    host
        The host of the WAGO Ethernet module.
    port
        The port serving the TCP modbus service.
    name
        The name associated with this WAGO controller.

    """

    def __init__(self, host: str, port: int = 502, name: str = ""):

        super().__init__(host, port)

        self.name = name

    async def read_sensors(
        self,
        units: bool = False,
    ) -> dict[str, float | tuple[float, str]]:
        """Read temperature and humidity sensors."""

        sensors = await self.read_category("temperature")
        rhs = await self.read_category("humidity")

        sensors.update(rhs)
        sensors = {k.split(".")[1].lower(): v for k, v in sensors.items()}

        if units:
            return sensors

        sensors_no_unit = {}
        for k, v in sensors.items():
            sensors_no_unit[k] = round(v[0], 2)

        return sensors_no_unit

    async def read_relays(self) -> dict[str, bool]:
        """Reads the status of the power relays.

        Returns
        -------
        power
            A dictionary with the status of the power relays. `True` means the
            relay is closed, `False` open.

        """

        relays = await self.read_category("relays")

        return {
            k.split(".")[1].lower(): True if v[0] == "closed" else False
            for k, v in relays.items()
        }

    async def set_relay(self, relay: str, closed: bool = True):
        """Sets the status of a power relay."""

        try:
            device = self.get_device(relay)
        except ValueError:
            raise NameError(f"Cannot find relay {relay!r}.")

        assert isinstance(device, Relay)

        status = await device.read()
        if status[0] == "closed" and closed is True:
            return None
        elif status[0] == "open" and closed is False:
            return None

        if closed:
            await device.close()
        else:
            await device.open()

        return True
